// Copyright (c) 2021 Tencent. All rights reserved.

#include "leb_network/audio_channel_receiver.h"

#include <algorithm>

#include "leb_network/converter/aac_adts_converter.h"
#include "leb_network/converter/opus_converter.h"
#include "leb_network/internal_defines.h"

namespace {

constexpr char kAacAdtsName[] = "MP4A-ADTS";
constexpr char kOpusName[] = "OPUS";

const int aac_sample_rates[16] = {
    96000, 88200, 64000, 48000, 44100, 32000,
    24000, 22050, 16000, 12000, 11025, 8000, 7350
};

bool CompareIngoreCase(const std::string& one, const std::string& other) {
  if (one.length() == other.length()) {
    return std::equal(
        one.begin(), one.end(), other.begin(),
        [](char lhs, char rhs) { return tolower(lhs) == tolower(rhs); });
  }

  return false;
}

}  // namespace

namespace liteav {
namespace lebnetwork {

AudioChannelReceiver::AudioChannelReceiver(
    std::shared_ptr<RTCPeerConnection> pc, LEBConnection::Client* client)
    : ChannelReceiver(pc, client), throtter_(60000),hasInformed_(false) {}

AudioChannelReceiver::~AudioChannelReceiver() = default;

// Implement RTCRtpReceiverDelegate
void AudioChannelReceiver::OnReceiveFirstPacket(RTCRtpMediaType mediaType) {}

int AudioChannelReceiver::getEndSeqNumOfLastDecodedFrame(
    RTCRtpMediaType mediaType, uint16_t* sequence_number, uint32_t* timestamp) {
  return -1;
}

void AudioChannelReceiver::OnReceiveAudioCodecs(
    const std::map<int, SdpAudioFormat>& codecs) {
  formats_ = codecs;
  client_->SetHasAudio();

  auto it = codecs.begin();
  while (it != codecs.end() && !hasInformed_) {
    AudioInfo info;
    if (CompareIngoreCase(it->second.name, kAacAdtsName)) {
      SdpAudioFormat format =  codecs.begin()->second;
      auto iter = format.parameters.find("config");
      if (iter != format.parameters.end()) {
        uint8_t header[1024];
        memcpy(header, iter->second.c_str(), iter->second.length());
        for (size_t i=0; i < iter->second.length(); i++) {
          if (header[i] >= '0' && header[i] <= '9')
            header[i] = header[i] - '0';
          else if (header[i] >= 'a' && header[i] <= 'f')
            header[i] = header[i] - 'a' + 10;
        }
        uint16_t bits = ((header[4]&0xf) << 12) | ((header[5]&0xf) << 8) | ((header[6]&0xf) << 4) | (header[7]&0xf);//remove first 15 bit
        bits >>= 1;
        info.codec_type = AudioCodecType::kAac;
        info.sample_rate = aac_sample_rates[(bits >> 7) & 0xf];
        info.num_channels = (bits >> 3) & 0xf;
        iter = format.parameters.find("SBR-enabled");
        if (iter != format.parameters.end()) {
            const char *str = iter->second.c_str();
            int sbr = atoi(str);
            if (sbr) {
              info.sample_rate <<= 1;
            }
        }
        iter = format.parameters.find("PS-enabled");
        if (iter != format.parameters.end()) {
            const char *str = iter->second.c_str();
            int ps = atoi(str);
            if (ps) {
              info.num_channels <<= 1;
            }
        }
      } else {
        info.codec_type = AudioCodecType::kAac;
        info.sample_rate = it->second.clockrate_hz;
        info.num_channels = it->second.num_channels;
      }
      NotifyLog("", LogLevel::kInfo, "sample_rate %d, num_channels %d", info.sample_rate, info.num_channels);
      client_->OnAudioInfo(info);
      hasInformed_ = true;
    } else if (CompareIngoreCase(it->second.name, kOpusName)) {
      info.codec_type = AudioCodecType::kOpus;
      info.sample_rate = it->second.clockrate_hz;
      info.num_channels = it->second.num_channels;
      client_->OnAudioInfo(info);
      hasInformed_ = true;
    } else {
      NotifyLog("", LogLevel::kWarning, "Unknown audio codec %s", it->second.name.c_str());
    }
    
    it++;
  }
}

void AudioChannelReceiver::OnReceiveVideoCodecs(
    const std::map<int, SdpVideoFormat>& codecs) {}

// Implement RtpPacketSinkInterface
void AudioChannelReceiver::OnRtpPacket(const RtpPacket& packet,
                                       RTCRtpMediaType mediaType) {
  EnsureConverterGenerated(packet.header.payloadType);
  if (converter_) {
    converter_->Append(packet);

    if (client_->HasAudio() && !client_->HasAudioInformed())
      return;

    std::vector<std::unique_ptr<EncodedAudioFrame>> frames =
        converter_->Convert();
    for (auto& frame : frames) {
      if (!hasInformed_) {
        AudioInfo info;
        info.codec_type = frame->codec_type;
        client_->OnAudioInfo(info);;
        hasInformed_ = true;
      }
      client_->OnEncodedAudio(std::move(frame));
    }
  }
}

void AudioChannelReceiver::OnRtpMetaDataPacket(const RtpMetaData& packet,
                                               RTCRtpMediaType mediaType) {}

void AudioChannelReceiver::OnMediaPacket(const void* data, int length,
                                         int64_t pts, RTCRtpMediaType mediaType,
                                         bool iscodec) {}

void AudioChannelReceiver::EnsureConverterGenerated(uint8_t payload_type) {
  if (converter_) {
    return;
  }

  auto iter = formats_.find(payload_type);
  if (iter != formats_.end()) {
    if (CompareIngoreCase(iter->second.name, kAacAdtsName)) {
      converter_.reset(new AacAdtsConverter(iter->second));
    } else if (CompareIngoreCase(iter->second.name, kOpusName)) {
      converter_.reset(new OpusConverter(iter->second));
    } else {
      if (throtter_.ShouldGo()) {
        NotifyLog("", LogLevel::kWarning, "Unknown audio format %s",
                  iter->second.name.c_str());
      }
    }
  } else {
    if (throtter_.ShouldGo()) {
      NotifyLog("", LogLevel::kWarning, "Unknown audio payload type %d",
                static_cast<int>(payload_type));
    }
  }
}

}  // namespace lebnetwork
}  // namespace liteav
