/*
 *  Copyright 2016 The WebRTC Project Authors. All rights reserved.
 *
 *  Use of this source code is governed by a BSD-style license
 *  that can be found in the LICENSE file in the root of the source
 *  tree. An additional intellectual property rights grant can be found
 *  in the file PATENTS.  All contributing project authors may
 *  be found in the AUTHORS file in the root of the source tree.
 */

#include "leb_network/base/copy_on_write_buffer.h"

#include <stddef.h>

namespace trtc {

CopyOnWriteBuffer::CopyOnWriteBuffer() : offset_(0), size_(0) {}

CopyOnWriteBuffer::CopyOnWriteBuffer(const CopyOnWriteBuffer& buf)
    : buffer_(buf.buffer_), offset_(buf.offset_), size_(buf.size_) {}

CopyOnWriteBuffer::CopyOnWriteBuffer(CopyOnWriteBuffer&& buf)
    : buffer_(std::move(buf.buffer_)), offset_(buf.offset_), size_(buf.size_) {
  buf.offset_ = 0;
  buf.size_ = 0;
}

CopyOnWriteBuffer::CopyOnWriteBuffer(const std::string& s)
    : CopyOnWriteBuffer(s.data(), s.length()) {}

CopyOnWriteBuffer::CopyOnWriteBuffer(size_t size)
    : buffer_(size > 0 ? std::make_shared<Buffer>() : nullptr),
      offset_(0),
      size_(size) {
  // defer SetSize
  if (size > 0) {
    buffer_->SetSize(size);
  }
}

CopyOnWriteBuffer::CopyOnWriteBuffer(size_t size, size_t capacity)
    : buffer_(size > 0 || capacity > 0 ? std::make_shared<Buffer>() : nullptr),
      offset_(0),
      size_(size) {
  // defer SetSize/EnsureCapacity
  if (size > 0) {
    buffer_->SetSize(size);
  }
  if (capacity > 0) {
    buffer_->EnsureCapacity(capacity);
  }
}

CopyOnWriteBuffer::~CopyOnWriteBuffer() = default;

bool CopyOnWriteBuffer::operator==(const CopyOnWriteBuffer& buf) const {
  // Must either be the same view of the same buffer or have the same contents.
  return size_ == buf.size_ &&
         (cdata() == buf.cdata() || memcmp(cdata(), buf.cdata(), size_) == 0);
}

void CopyOnWriteBuffer::SetSize(size_t size) {
  if (!buffer_) {
    if (size > 0) {
      buffer_ = std::make_shared<Buffer>();
      buffer_->SetSize(size);
      offset_ = 0;
      size_ = size;
    }
    return;
  }

  if (size <= size_) {
    size_ = size;
    return;
  }

  UnshareAndEnsureCapacity(std::max(capacity(), size));
  buffer_->SetSize(size + offset_);
  size_ = size;
}

void CopyOnWriteBuffer::EnsureCapacity(size_t new_capacity) {
  if (!buffer_) {
    if (new_capacity > 0) {
      buffer_ = std::make_shared<Buffer>();
      buffer_->EnsureCapacity(new_capacity);

      offset_ = 0;
      size_ = 0;
    }
    return;
  } else if (new_capacity <= capacity()) {
    return;
  }

  UnshareAndEnsureCapacity(new_capacity);
}

void CopyOnWriteBuffer::Clear() {
  if (!buffer_) return;

  if (buffer_) {
    buffer_->Clear();
  } else {
    size_t cap = capacity();
    buffer_ = std::make_shared<Buffer>();
    buffer_->EnsureCapacity(cap);
  }
  offset_ = 0;
  size_ = 0;
}

void CopyOnWriteBuffer::UnshareAndEnsureCapacity(size_t new_capacity) {
  if (buffer_ && new_capacity <= capacity()) {
    return;
  }

  auto tmp = std::make_shared<Buffer>();
  tmp->EnsureCapacity(new_capacity);
  tmp->SetData(buffer_->data() + offset_, size_);
  buffer_ = tmp;

  offset_ = 0;
}

}  // namespace trtc
