// Copyright (c) 2021 Tencent. All rights reserved.

#ifndef LEB_NETWORK_BASE_UNWRAPPER_H_
#define LEB_NETWORK_BASE_UNWRAPPER_H_

#include <stdint.h>

namespace liteav {

class Unwrapper {
 public:
  Unwrapper(uint64_t min_wrap, uint64_t max_wrap)
      : has_last_value_(false), keep_increase(false) {
    min_wrap_ = min_wrap;
    max_wrap_ = max_wrap;
    wrap_interval_ = max_wrap_ - min_wrap_ + 1;
    half_wrap_interval_ = wrap_interval_ >> 1;
  }

  uint64_t wrap(uint64_t unwrap_value) {
    has_last_value_ = true;
    last_unwrapped_ = unwrap_value;
    last_value_ = unwrap_value % wrap_interval_;
    return last_value_ + min_wrap_;
  }

  uint64_t Unwrap(uint64_t value) {
    if (value > max_wrap_ || value < min_wrap_) {
      return 0;
    }

    value -= min_wrap_;

    if (!has_last_value_) {
      has_last_value_ = true;
      last_value_ = value;
      if (keep_increase) {
        last_unwrapped_ = value /*+ wrap_interval_*/ +
                          last_unwrapped_ / wrap_interval_ * wrap_interval_;
      } else {
        last_unwrapped_ = value /*+ wrap_interval_*/;
      }
      return last_unwrapped_;
    }

    uint64_t unwrapped = 0;
    if (value >= last_value_) {
      uint64_t diff = value - last_value_;
      if (diff < half_wrap_interval_) {
        unwrapped = last_unwrapped_ + diff;
      } else {
        unwrapped = last_unwrapped_ - (wrap_interval_ - diff);
        if (last_unwrapped_ < (wrap_interval_ - diff)) {
          // Unwrap error!
        }
      }
    } else {
      uint64_t diff = last_value_ - value;
      if (diff < half_wrap_interval_) {
        unwrapped = last_unwrapped_ - diff;
        if (last_unwrapped_ < diff) {
          // Unwrap error!
        }
      } else {
        unwrapped = last_unwrapped_ + (wrap_interval_ - diff);
      }
    }

    last_value_ = value;
    last_unwrapped_ = unwrapped;
    return last_unwrapped_;
  }

  void reset(bool keepIncrease = false) {
    has_last_value_ = false;
    keep_increase = keepIncrease;
  }

 private:
  uint64_t last_unwrapped_;
  uint64_t last_value_;
  bool has_last_value_;
  uint64_t min_wrap_;
  uint64_t max_wrap_;
  uint64_t wrap_interval_;
  uint64_t half_wrap_interval_;
  bool keep_increase;
};

}  // namespace liteav

#endif  // LEB_NETWORK_BASE_UNWRAPPER_H_
