// Copyright (c) 2021 Tencent. All rights reserved.

#include "leb_network/converter/aac_adts_converter.h"

#include <stdint.h>

#include "leb_network/internal_defines.h"

namespace {

constexpr int kAdtsHeaderSize = 7;
constexpr int kCrcSize = 2;
constexpr int kSampleRates[] = {96000, 88200, 64000, 48000, 44100, 32000, 24000,
                                22050, 16000, 12000, 11025, 8000,  7350};

int GetSampleRateIndex(int sample_rate) {
  int size = sizeof(kSampleRates) / sizeof(kSampleRates[0]);
  for (int i = 0; i < size; ++i) {
    if (kSampleRates[i] == sample_rate) {
      return i;
    }
  }
  return -1;
}

int GetSampleRate(int sample_rate_index) {
  int size = sizeof(kSampleRates) / sizeof(kSampleRates[0]);
  if (sample_rate_index < size) {
    return kSampleRates[sample_rate_index];
  }
  return -1;
}

uint16_t GenerateSyncWord(int sample_rate, int channels) {
  uint16_t audio_specific_config = 0;
  char buf_position = 16;

  // AAC-LC
  buf_position -= 5;
  audio_specific_config |= ((2 << buf_position) & 0xF800);

  // sample rate index
  buf_position -= 4;
  int sample_rate_index = GetSampleRateIndex(sample_rate);
  audio_specific_config |= ((sample_rate_index << buf_position) & 0x0780);

  // channels
  buf_position -= 4;
  audio_specific_config |= ((channels << buf_position) & 0x78);

  // other
  audio_specific_config |= (0 & 0x07);

  uint16_t rst = 0;
  uint8_t* p_rst = reinterpret_cast<uint8_t*>(&rst);
  p_rst[0] = (audio_specific_config >> 8) & 0xFF;
  p_rst[1] = (audio_specific_config & 0xFF);
  return rst;
}

}  // namespace

namespace liteav {
namespace lebnetwork {

AacAdtsConverter::AacAdtsConverter(const SdpAudioFormat& format)
    : format_(format), sync_word_generated_(false),
      unwrapper_dts_(0, std::numeric_limits<uint32_t>::max()) {}

AacAdtsConverter::~AacAdtsConverter() = default;

void AacAdtsConverter::Append(const RtpPacket& packet) {
#if 0 //use adif frame
  if (!sync_word_generated_) {
    sync_word_generated_ = true;
    auto data = packet.payload();
    int sample_rate_index = (data[2] & 0x3c) >> 2;
    int sample_rate = GetSampleRate(sample_rate_index);
    int channels = ((data[2] & 0x01) << 2) | ((data[3] & 0xC0) >> 6);
    // Add ADIF sync word
    // HE-AAC 的真实采样率是 ADTS 头部记录的采样率的 2 倍，所以这里计算 ADIF
    // 同步字时，不能直接使用 format_.clockrate_hz，而应该使用 ADTS 头里的采样率
    uint16_t sync_word = GenerateSyncWord(sample_rate, channels);
    uint8_t* p_sync_word = reinterpret_cast<uint8_t*>(&sync_word);

    std::unique_ptr<EncodedAudioFrame> frame(new EncodedAudioFrame());

    CopyAudioInfo(packet, format_, frame.get());
    frame->data.SetData(p_sync_word, 2);
    frame->frame_type = AudioFrameType::kAacAdifHeader;
    frame->output_delay_ms = pkt.output_delay_ms;
  
    frames_.push_back(std::move(frame));
  }

  // ADTS format see https://www.cnblogs.com/zhangxuan/p/8809245.html
  // Remove ADTS header
  auto data = packet.payload();
  int header_size = kAdtsHeaderSize;
  if (!(data[1] & 0x1)) {
    header_size += kCrcSize;
  }
  std::unique_ptr<EncodedAudioFrame> frame(new EncodedAudioFrame());

  CopyAudioInfo(packet, format_, frame.get());
  frame->data.SetData(data + header_size, packet.payload_size() - header_size);
  frame->frame_type = AudioFrameType::kAacAdifData;
#else // use adts frame
  std::unique_ptr<EncodedAudioFrame> frame(new EncodedAudioFrame());

  // parse adts header
  auto data = packet.payload();
  int sample_rate_index = (data[2] & 0x3c) >> 2;
  int sample_rate = GetSampleRate(sample_rate_index);
  int channels = ((data[2] & 0x01) << 2) | ((data[3] & 0xC0) >> 6);

  frame->codec_type = AudioCodecType::kAac;
  frame->bits_per_sample = 16;
  frame->num_channels = channels;
  frame->pts = (int64_t)unwrapper_dts_.Unwrap(static_cast<uint64_t>(packet.decoding_timestamp));
  frame->sample_rate = sample_rate;
  frame->data.SetData(data, packet.payload_size());
  frame->frame_type = AudioFrameType::kAacAdts;
  frame->output_delay_ms = packet.output_delay_ms;
#endif
  frames_.push_back(std::move(frame));
}

std::vector<std::unique_ptr<EncodedAudioFrame>> AacAdtsConverter::Convert() {
  return std::move(frames_);
}

}  // namespace lebnetwork
}  // namespace liteav
