// Copyright (c) 2021 Tencent. All rights reserved.

#include "leb_network/converter/h264_converter.h"

#include <stdint.h>

#include "leb_network/internal_defines.h"

namespace {

constexpr uint8_t kNalTypeBits = 0x1F;

constexpr uint8_t kNalStartCode[] = {0x00, 0x00, 0x00, 0x01};

uint8_t NalType(const uint8_t nal_head) { return nal_head & kNalTypeBits; }

bool IsKeyFrame(const uint8_t nal_head) {
  uint8_t nalu_type = NalType(nal_head);
  return nalu_type == liteav::lebnetwork::NaluType::kIdr;
}

bool IsSpsOrPps(const uint8_t nal_head) {
  uint8_t nalu_type = NalType(nal_head);
  if (nalu_type == liteav::lebnetwork::NaluType::kSps ||
      nalu_type == liteav::lebnetwork::NaluType::kPps) {
    return true;
  }
  return false;
}

}  // namespace

namespace liteav {
namespace lebnetwork {

H264Converter::H264Converter()
    : unwrapper_seqnum_(0, std::numeric_limits<uint16_t>::max()),
      unwrapper_dts_(0, std::numeric_limits<uint32_t>::max()),
      is_key_frame_(false),
      nal_head_(0),
      throtter_(60000) {}

H264Converter::~H264Converter() = default;

void H264Converter::Append(const RtpPacket& packet) {
  uint8_t* payload = const_cast<uint8_t*>(packet.payload());

  // FU Indicator 后五位即 NALU 类型 0x1F = 0001 1111
  uint8_t nalu_type = NalType(payload[0]);
  if (nalu_type < NaluType::kStapA) {
    // NaluType 为 1-23  的 rtp 包有且只有一个 NAL 包
    nal_buffer_.AppendData(kNalStartCode, sizeof(kNalStartCode));
    nal_buffer_.AppendData(payload, packet.payload_size());

    is_key_frame_ = IsKeyFrame(payload[0]);
  } else if (nalu_type == NaluType::kStapA) {
    StapAToAnnexB(payload, packet.payload_size(), &nal_buffer_, &is_key_frame_);
  } else if (nalu_type == NaluType::kFuA) {
    FuAToAnnexB(payload, packet.payload_size(), &nal_buffer_, &is_key_frame_);
  } else {
    // Unsupported
    if (throtter_.ShouldGo()) {
      NotifyLog("", LogLevel::kWarning, "Unsupported Nal unit type %d",
                static_cast<int>(nalu_type));
    }
  }

  if (packet.is_last_packet_in_frame && nal_buffer_.size()) {
    const media_transport::VCMPacket& pkt =
        static_cast<const media_transport::VCMPacket&>(packet);
    std::unique_ptr<EncodedVideoFrame> frame(new EncodedVideoFrame());
    frame->codec_type = VideoCodecType::kH264;
    frame->dts = (int64_t)unwrapper_dts_.Unwrap(static_cast<uint64_t>(pkt.decoding_timestamp));
    frame->pts = frame->dts + pkt.video_composition_time;
    frame->data = std::move(nal_buffer_);
    frame->is_key_frame = is_key_frame_;
    frame->sequence_number =
        unwrapper_seqnum_.Unwrap(static_cast<uint64_t>(pkt.header.sequenceNumber));
    frame->output_delay_ms = pkt.output_delay_ms;

    frames_.push_back(std::move(frame));
  }
}

std::vector<std::unique_ptr<EncodedVideoFrame>> H264Converter::Convert() {
  return std::move(frames_);
}

bool H264Converter::StapAToAnnexB(const uint8_t* stapa, size_t size,
                                  trtc::CopyOnWriteBuffer* annexb,
                                  bool* is_key_frame) {
  // Stap-A format: stap-a nal header(1 byte) + nal size(2 bytes) +
  // nal data(dynamic) + nal size(2 bytes) + nal data(dynamic) + ...
  // see https://www.cnblogs.com/jixiaohua/p/11899440.html
  int32_t index = 1;  // skip 1 byte header
  bool contain_video_data = false;
  while (index < static_cast<int32_t>(size)) {
    uint16_t nal_size = static_cast<uint16_t>(stapa[index]);
    nal_size = (nal_size << 8) & stapa[index + 1];
    index += 2;
    int32_t next_index = index + nal_size;
    if (next_index > static_cast<int32_t>(size)) {
      // Overflow, data invalid
      NotifyLog("", LogLevel::kWarning, "Invalid Nal data");
      return contain_video_data;
    }

    if (!IsSpsOrPps(stapa[index])) {
      contain_video_data = true;
      *is_key_frame = IsKeyFrame(stapa[index]);
    }

    annexb->AppendData(kNalStartCode, sizeof(kNalStartCode));
    annexb->AppendData(stapa + index, nal_size);

    index = next_index;
  }

  return contain_video_data;
}

bool H264Converter::FuAToAnnexB(const uint8_t* fua, size_t size,
                                trtc::CopyOnWriteBuffer* annexb,
                                bool* is_key_frame) {
  // FU-A format: FU-A header(2 bytes) + FU payload(dynamic)
  // The first two bytes are the FU indicator and the FU header

  // FU-A Header start 标记位  0x80 = 1000 0000
  uint8_t start_bit = fua[1] & 0x80;
  // FU-A Header End 标记位    0x40= 0100 0000
  uint8_t end_bit = fua[1] & 0x40;

  if (start_bit) {
    annexb->AppendData(kNalStartCode, sizeof(kNalStartCode));
    // Recover 1 byte nal header
    nal_head_ = (fua[0] & 0xE0) | (fua[1] & kNalTypeBits);
    annexb->AppendData(&nal_head_, 1);

    *is_key_frame = IsKeyFrame(nal_head_);
  }

  // Copy payload
  annexb->AppendData(fua + 2, size - 2);

  return !IsSpsOrPps(nal_head_) && end_bit;
}

}  // namespace lebnetwork
}  // namespace liteav
