// Copyright (c) 2021 Tencent. All rights reserved.

#include "leb_network/converter/h265_converter.h"

#include <stdint.h>

#include <string>

#include "leb_network/internal_defines.h"

namespace {

constexpr uint8_t kNalTypeBits = 0x3F;

constexpr uint8_t kRtpHevcPayloadHeaderSize = 2;

constexpr uint8_t kRtpHevcDonlFieldSize = 2;
constexpr uint8_t kRtpHevcDondFieldSize = 1;

constexpr char kSpropMaxDonDiff[] = "sprop-max-don-diff";
constexpr char kSpropDepackBufNalus[] = "sprop-depack-buf-nalus";

constexpr uint8_t kNalStartCode[] = {0x00, 0x00, 0x00, 0x01};

uint8_t NalType(const uint8_t nal_head) {
  return (nal_head >> 1) & kNalTypeBits;
}

bool IsKeyFrame(const uint8_t nal_head) {
  uint8_t nalu_type = NalType(nal_head);
  return nalu_type == liteav::lebnetwork::H265NaluType::kH265Idr ||
         nalu_type == liteav::lebnetwork::H265NaluType::kH265Cra;
}

bool IsVpsOrSpsOrPps(const uint8_t nal_head) {
  uint8_t nalu_type = NalType(nal_head);
  if (nalu_type == liteav::lebnetwork::H265NaluType::kH265Vps ||
      nalu_type == liteav::lebnetwork::H265NaluType::kH265Sps ||
      nalu_type == liteav::lebnetwork::H265NaluType::kH265Pps) {
    return true;
  }
  return false;
}

bool IsNumeric(std::string str) {
  for (auto v : str) {
    if (v < '0' || v > '9') {
      return false;
    }
  }
  return true;
}

}  // namespace

namespace liteav {
namespace lebnetwork {

H265Converter::H265Converter(const SdpVideoFormat& format)
    : unwrapper_seqnum_(0, std::numeric_limits<uint16_t>::max()),
      unwrapper_dts_(0, std::numeric_limits<uint32_t>::max()),
      is_key_frame_(false),
      throtter_(60000) {
  nal_head_[0] = 0;
  nal_head_[1] = 0;
  auto v = format.parameters.find(kSpropMaxDonDiff);
  if (v != format.parameters.end() && IsNumeric(v->second)) {
    int number = std::stoi(v->second);
    if (number != 0) {
      using_donl_field_ = true;
    }
  }

  v = format.parameters.find(kSpropDepackBufNalus);
  if (v != format.parameters.end() && IsNumeric(v->second)) {
    int number = std::stoi(v->second);
    if (number != 0) {
      using_donl_field_ = true;
    }
  }
}

H265Converter::~H265Converter() = default;

void H265Converter::Append(const RtpPacket& packet) {
  /*
   * decode the HEVC payload header according to section 4 of draft version 6:
   *
   *    0                   1
   *    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
   *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   *   |F|   Type    |  LayerId  | TID |
   *   +-------------+-----------------+
   *
   *      Forbidden zero (F): 1 bit
   *      NAL unit type (Type): 6 bits
   *      NUH layer ID (LayerId): 6 bits
   *      NUH temporal ID plus 1 (TID): 3 bits
   */
  // reference to rtpdec_hevc.c in FFmpeg libavformat
  uint8_t* payload = const_cast<uint8_t*>(packet.payload());

  uint8_t nalu_type = NalType(payload[0]);
  if (nalu_type < H265NaluType::kH265Ap) {
    // 单个 NAL 包
    //   0                   1                   2                   3
    //   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //  |           PayloadHdr          |      DONL (conditional)       |
    //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //  |                                                               |
    //  |                  NAL unit payload data                        |
    //  |                                                               |
    //  |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    //  |                               :...OPTIONAL RTP padding        |
    //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    nal_buffer_.AppendData(kNalStartCode, sizeof(kNalStartCode));
    nal_buffer_.AppendData(payload, packet.payload_size());

    is_key_frame_ = IsKeyFrame(payload[0]);
  } else if (nalu_type == H265NaluType::kH265Ap) {
    ApToAnnexB(payload, packet.payload_size(), &nal_buffer_, &is_key_frame_);
  } else if (nalu_type == H265NaluType::kH265Fu) {
    FuToAnnexB(payload, packet.payload_size(), &nal_buffer_, &is_key_frame_);
  } else {
    // Unsupported
    if (throtter_.ShouldGo()) {
      NotifyLog("", LogLevel::kWarning, "Unsupported H265 Nal unit type %d",
                static_cast<int>(nalu_type));
    }
  }

  if (packet.is_last_packet_in_frame && nal_buffer_.size()) {
    const media_transport::VCMPacket& pkt =
        static_cast<const media_transport::VCMPacket&>(packet);
    std::unique_ptr<EncodedVideoFrame> frame(new EncodedVideoFrame());
    frame->codec_type = VideoCodecType::kH265;
    frame->dts = (int64_t)unwrapper_dts_.Unwrap(static_cast<uint64_t>(pkt.decoding_timestamp));
    frame->pts = frame->dts + pkt.video_composition_time;
    frame->data = std::move(nal_buffer_);
    frame->is_key_frame = is_key_frame_;
    frame->sequence_number =
        unwrapper_seqnum_.Unwrap(static_cast<uint64_t>(pkt.header.sequenceNumber));
    frame->output_delay_ms = pkt.output_delay_ms;

    frames_.push_back(std::move(frame));
  }
}

std::vector<std::unique_ptr<EncodedVideoFrame>> H265Converter::Convert() {
  return std::move(frames_);
}

// Agregated packet (AP) -> AnnexB
bool H265Converter::ApToAnnexB(const uint8_t* ap, size_t size,
                               trtc::CopyOnWriteBuffer* annexb,
                               bool* is_key_frame) {
  //   0                   1                   2                   3
  //   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
  //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //  |   PayloadHdr (Type=48)        |        NALU 1 DONL            |
  //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //  |          NALU 1 Size          |            NALU 1 HDR         |
  //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //  |                                                               |
  //  |                 NALU 1 Data   . . .                           |
  //  |                                                               |
  //  +     . . .     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //  |               |  NALU 2 DOND  |          NALU 2 Size          |
  //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //  |          NALU 2 HDR           |                               |
  //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+          NALU 2 Data          |
  //  |                                                               |
  //  |        . . .                  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //  |                               :...OPTIONAL RTP padding        |
  //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  int32_t index = kRtpHevcPayloadHeaderSize;  // skip 2 byte header
  bool contain_video_data = false;
  int32_t skip_dond_field_size = 0;
  if (using_donl_field_) {
    index += kRtpHevcDonlFieldSize;
    skip_dond_field_size = kRtpHevcDondFieldSize;
  }
  while (index < static_cast<int32_t>(size)) {
    uint16_t nal_size = static_cast<uint16_t>(ap[index]);
    nal_size = (nal_size << 8) & ap[index + 1];
    index += 2;
    int32_t next_index = index + nal_size;
    if (next_index > static_cast<int32_t>(size)) {
      // Overflow, data invalid
      NotifyLog("", LogLevel::kWarning, "Invalid Nal data");
      return contain_video_data;
    }

    if (!IsVpsOrSpsOrPps(ap[index])) {
      contain_video_data = true;
      *is_key_frame = IsKeyFrame(ap[index]);
    }

    annexb->AppendData(kNalStartCode, sizeof(kNalStartCode));
    annexb->AppendData(ap + index, nal_size);

    index = next_index;
    index += skip_dond_field_size;
  }

  return contain_video_data;
}

//
bool H265Converter::FuToAnnexB(const uint8_t* fu, size_t size,
                               trtc::CopyOnWriteBuffer* annexb,
                               bool* is_key_frame) {
  //   0                   1                   2                   3
  //   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
  //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //  |    PayloadHdr (Type=49)       |   FU header   | DONL (cond)   |
  //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  //  | DONL (cond)   |                                               |
  //  |-+-+-+-+-+-+-+-+                                               |
  //  |                         FU payload                            |
  //  |                                                               |
  //  |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  //  |                               :...OPTIONAL RTP padding        |
  //  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  // FU header:
  //  +---------------+
  //  |0|1|2|3|4|5|6|7|
  //  +-+-+-+-+-+-+-+-+
  //  |S|E|  FuType   |
  //  +---------------+
  uint8_t start_bit = fu[kRtpHevcPayloadHeaderSize] & 0x80;
  uint8_t end_bit = fu[kRtpHevcPayloadHeaderSize] & 0x40;

  if (start_bit) {
    annexb->AppendData(kNalStartCode, sizeof(kNalStartCode));
    // Recover 1 byte nal header
    nal_head_[0] =
        (fu[0] & 0x81) | ((fu[kRtpHevcPayloadHeaderSize] & 0x3f) << 1);
    nal_head_[1] = fu[1];
    annexb->AppendData(nal_head_, 2);

    *is_key_frame = IsKeyFrame(nal_head_[0]);
  }

  // Copy payload
  fu += kRtpHevcPayloadHeaderSize;
  size -= kRtpHevcPayloadHeaderSize;
  fu += 1;
  size -= 1;
  if (using_donl_field_) {
    fu += kRtpHevcDonlFieldSize;
    size -= kRtpHevcDonlFieldSize;
  }
  annexb->AppendData(fu, size);

  return !IsVpsOrSpsOrPps(nal_head_[0]) && end_bit;
}

}  // namespace lebnetwork
}  // namespace liteav
