// Copyright (c) 2021 Tencent. All rights reserved.

#include "leb_network/leb_connection_impl.h"

#include "leb_network/audio_channel_receiver.h"
#include "leb_network/internal_defines.h"
#include "leb_network/leb_defines.h"
#include "leb_network/metadata_channel_receiver.h"
#include "leb_network/video_channel_receiver.h"
// #include "txg_time_util.h"

namespace {

liteav::lebnetwork::Stats RTCStatsReportToStats(const RTCStatsReport& stats) {
  liteav::lebnetwork::Stats st;
  st.first_audio_received_cost_ms = stats.first_audio_frame_output_delay_ms;
  st.first_video_received_cost_ms = stats.first_video_frame_output_delay_ms;
  st.rtt_ms = stats.rttInMs;
  st.audio_bytes_received = stats.audioBytesReceived;
  st.audio_nack_count = stats.audioNackCount;
  st.audio_packets_lost = stats.audioPacketsLost;
  st.audio_packets_received = stats.audioPacketsReceived;
  st.video_nack_count = stats.videoNackCount;
  st.video_bytes_received = stats.videoBytesReceived;
  st.video_packets_lost = stats.videoPacketsLost;
  st.video_packets_received = stats.videoPacketsReceived;
  return st;
}

}  // namespace

namespace liteav {
namespace lebnetwork {

LEBConnectionImpl::LEBConnectionImpl(LEBConnection::Client* client)
    : client_(client),
      pc_delegate_hander_(new PeerConnectionDelegateHandler(this)),
      start_time_(0),
      signal_dns_cost_ms_(0),
      first_video_received_time_(0),
      first_audio_received_time_(0) {}

LEBConnectionImpl::~LEBConnectionImpl() { Stop(); }

bool LEBConnectionImpl::Start(const Configuration& config) {
  if (pc_) {
    NotifyLog("", LogLevel::kWarning,
              "LEBConnection Start() failed. Call Stop() first!");
    return false;
  }

  //  start_time_ = txf_gettickcount();
  first_video_received_time_ = 0;
  first_audio_received_time_ = 0;

  media_transport::RTCConfiguration cfg;
  cfg.enableAAC = config.enable_aac;
  cfg.disableEncryption = !config.enable_encryption;
  cfg.receiveAudio = config.receive_audio;
  cfg.receiveVideo = config.receive_video;
  cfg.enableFlexFec = config.enable_flex_fec;
  cfg.enableUdpSignal = config.enable_udp_signal;
  cfg.stats_period_ms = config.stats_period_ms;
  cfg.enableAudioPlc = config.enable_audio_plc;
  cfg.enable0Rtt = config.enable_0rtt;
  cfg.min_delay_ms = config.min_delay_ms;
  cfg.remoteAddress = config.remote_address;
  cfg.remotePort = 8000;
  cfg.streamUrl = config.stream_url;

  pc_.reset(new media_transport::RTCPeerConnection());
  pc_->setConfiguration(cfg);
  pc_->setDelegate(pc_delegate_hander_);

  sdp_exchange_.reset(new SessionDescriptionExchange(pc_, this, this));
  sdp_exchange_->CreateOffer();
  return true;
}

void LEBConnectionImpl::Stop() {
  if (!pc_) {
    return;
  }
  sdp_exchange_.reset();
  pc_->stopPeerConnection();
  pc_->closePeerConnection();
  pc_.reset();
  start_time_ = 0;
}

void LEBConnectionImpl::GetStats() {
  if (pc_) {
    pc_->getStats();
  }
}

void LEBConnectionImpl::SetRemoteSDP(const SdpInfo& info) {
  if (!pc_ || !sdp_exchange_) {
    return;
  }
  RTCSessionDescription remote_sdp;
  remote_sdp.type = RTCSdpType::RTCSdpTypeAnswer;
  remote_sdp.sdp = info.sdp;
  sdp_exchange_->SetRemoteSDP(remote_sdp);
}

// Implement RTCPeerConnectionDelegate
void LEBConnectionImpl::onChangeConnectionState(RTCPeerConnection* pc,
                                                RTCPeerConnectionState state) {
  if (pc_.get() != pc) {
    return;
  }

  switch (state) {
    case RTCPeerConnectionState::RTCPeerConnectionStateConnecting:
      OnNetState(NetState::kConnecting);
      break;
    case RTCPeerConnectionState::RTCPeerConnectionStateConnected:
      OnNetState(NetState::kConnected);
      break;
    case RTCPeerConnectionState::RTCPeerConnectionStateDisconnected:
      OnNetState(NetState::kDisconnected);
      break;
    case RTCPeerConnectionState::RTCPeerConnectionStateFailed:
      OnNetState(NetState::kConnectFailed);
      break;
    case RTCPeerConnectionState::RTCPeerConnectionStateClosed:
      OnNetState(NetState::kConnectClosed);
      break;
    default:
      break;
  }
}

void LEBConnectionImpl::onAddReceiver(
    RTCPeerConnection* pc, std::shared_ptr<RTCRtpReceiver> receiver) {
  if (pc_.get() != pc) {
    return;
  }

  std::shared_ptr<ChannelReceiver> channel_receiver;
  switch (receiver->getMediaType()) {
    case RTCRtpMediaType::RTCRtpMediaTypeAudio:
      channel_receiver.reset(new AudioChannelReceiver(pc_, this));
      break;
    case RTCRtpMediaType::RTCRtpMediaTypeVideo:
      channel_receiver.reset(new VideoChannelReceiver(pc_, this));
      break;
#ifdef WEBRTC_BUILD_DATA_CHANNEL
    case RTCRtpMediaType::RTCRtpMediaTypeData:
      channel_receiver.reset(new MetadataChannelReceiver(pc_, this));
      break;
#endif
  }
  receiver->setDelegate(channel_receiver);
  receiver->setSink(channel_receiver);

  receivers_.insert(
      std::pair<RTCRtpMediaType, std::shared_ptr<ChannelReceiver>>(
          receiver->getMediaType(), channel_receiver));
}

void LEBConnectionImpl::onChangeCandidate(RTCPeerConnection* pc,
                                          RTCIceCandidate& local,
                                          RTCIceCandidate& remote,
                                          int last_data_received_ms,
                                          std::string reason) {
  if (pc_.get() != pc) {
    return;
  }

  data_server_ip_ = remote.serverIp;
}

void LEBConnectionImpl::onStatsReport(RTCStatsReport& stats) {
  Stats st = RTCStatsReportToStats(stats);
  //  uint64_t now_ms = txf_getnstickcount();
  //  st.connect_cost_ms =
  //      now_ms > start_time_ ? static_cast<int32_t>(now_ms - start_time_) : 0;
  st.data_server_ip = data_server_ip_;
  st.signal_server_ip = signal_server_ip_;
  st.signal_dns_cost_ms = static_cast<int32_t>(signal_dns_cost_ms_);
  OnStats(st);
}

void LEBConnectionImpl::onError(RTCPeerConnection* pc, const RTCError* error) {
  if (pc_.get() != pc) {
    return;
  }

  Error e = RTCErrorToError(*error);
  OnError(e);
}

// Implement SessionDescriptionExchange::SDPExchangeObserver
void LEBConnectionImpl::OnOfferCreated(RTCSessionDescription& sdp) {
  SdpInfo info;
  info.sdp_type = sdp.type == RTCSdpType::RTCSdpTypeOffer ? SdpType::kOffer : SdpType::kAnswer;
  info.sdp = sdp.sdp;
  client_->OnSdpInfo(info);
}

void LEBConnectionImpl::OnSignaling(const std::string& server_ip,
                                    const int64_t dns_cost_ms) {
  signal_server_ip_ = server_ip;
  signal_dns_cost_ms_ = dns_cost_ms;
}

void LEBConnectionImpl::OnVideoInfo(VideoInfo& info) {
  client_->OnVideoInfo(info);
  hasVideoInformed = true;
}

void LEBConnectionImpl::OnAudioInfo(AudioInfo& info) {
  client_->OnAudioInfo(info);
  hasAudioInformed = true;
}

void LEBConnectionImpl::OnEncodedVideo(
    std::unique_ptr<EncodedVideoFrame> frame) {
  //  if (!first_video_received_time_) {
  //    first_video_received_time_ = txf_getnstickcount();
  //  }
  client_->OnEncodedVideo(std::move(frame));
}

void LEBConnectionImpl::OnEncodedAudio(
    std::unique_ptr<EncodedAudioFrame> frame) {
  //  if (!first_audio_received_time_) {
  //    first_audio_received_time_ = txf_getnstickcount();
  //  }
  client_->OnEncodedAudio(std::move(frame));
}

void LEBConnectionImpl::OnMetadata(std::unique_ptr<Metadata> data) {
  client_->OnMetadata(std::move(data));
}

void LEBConnectionImpl::OnError(const Error& error) { client_->OnError(error); }

void LEBConnectionImpl::OnNetState(NetState state) {
  client_->OnNetState(state);
}

void LEBConnectionImpl::OnStats(const Stats& stats) { client_->OnStats(stats); }

LEBConnectionImpl::PeerConnectionDelegateHandler::PeerConnectionDelegateHandler(
    LEBConnectionImpl* pc_impl)
    : pc_impl_(pc_impl) {}

LEBConnectionImpl::PeerConnectionDelegateHandler::
    ~PeerConnectionDelegateHandler() = default;

// Implement RTCPeerConnectionDelegate
void LEBConnectionImpl::PeerConnectionDelegateHandler::onChangeConnectionState(
    RTCPeerConnection* pc, RTCPeerConnectionState state) {
  pc_impl_->onChangeConnectionState(pc, state);
}

void LEBConnectionImpl::PeerConnectionDelegateHandler::onAddReceiver(
    RTCPeerConnection* pc, std::shared_ptr<RTCRtpReceiver> receiver) {
  pc_impl_->onAddReceiver(pc, receiver);
}

void LEBConnectionImpl::PeerConnectionDelegateHandler::onChangeCandidate(
    RTCPeerConnection* pc, RTCIceCandidate& local, RTCIceCandidate& remote,
    int last_data_received_ms, std::string reason) {
  pc_impl_->onChangeCandidate(pc, local, remote, last_data_received_ms, reason);
}

void LEBConnectionImpl::PeerConnectionDelegateHandler::onStatsReport(
    RTCStatsReport& stats) {
  pc_impl_->onStatsReport(stats);
}

void LEBConnectionImpl::PeerConnectionDelegateHandler::onError(
    RTCPeerConnection* pc, const RTCError* error) {
  pc_impl_->onError(pc, error);
}

}  // namespace lebnetwork
}  // namespace liteav
