// Copyright (c) 2021 Tencent. All rights reserved.

#ifndef LEB_NETWORK_LEB_CONNECTION_IMPL_H_
#define LEB_NETWORK_LEB_CONNECTION_IMPL_H_

#include <map>

#include "include/peer_connection.h"
#include "leb_network/channel_receiver.h"
#include "leb_network/internal_defines.h"
#include "leb_network/leb_connection.h"
#include "leb_network/session_description_exchange.h"

namespace liteav {
namespace lebnetwork {

class LEBConnectionImpl
    : public LEBConnection,
      public LEBConnection::Client,
      public RTCPeerConnectionDelegate,
      public SessionDescriptionExchange::SDPExchangeObserver {
 public:
  ~LEBConnectionImpl();

  bool Start(const Configuration& config) override;

  // Used only for http signal
  void SetRemoteSDP(const SdpInfo& info) override;

  void Stop() override;

  void GetStats() override;

  // Implement RTCPeerConnectionDelegate
  void onChangeConnectionState(RTCPeerConnection* pc,
                               RTCPeerConnectionState state) override;

  void onAddReceiver(RTCPeerConnection* pc,
                     std::shared_ptr<RTCRtpReceiver> receiver) override;

  void onChangeCandidate(RTCPeerConnection* pc, RTCIceCandidate& local,
                         RTCIceCandidate& remote, int last_data_received_ms,
                         std::string reason) override;

  void onStatsReport(RTCStatsReport& stats) override;

  void onError(RTCPeerConnection* pc, const RTCError* error) override;

  // Implement SessionDescriptionExchange::SDPExchangeObserver
  void OnOfferCreated(RTCSessionDescription& sdp) override;
  void OnSignaling(const std::string& server_ip,
                   const int64_t dns_cost_ms) override;

  // Implement LEBConnection::Client
  void OnSdpInfo(SdpInfo& info) override {}
  void OnVideoInfo(VideoInfo& info) override;
  void OnAudioInfo(AudioInfo& info) override;
  void OnEncodedVideo(std::unique_ptr<EncodedVideoFrame> frame) override;
  void OnEncodedAudio(std::unique_ptr<EncodedAudioFrame> frame) override;
  void OnMetadata(std::unique_ptr<Metadata> data) override;
  void OnError(const Error& error) override;
  void OnNetState(NetState state) override;
  void OnStats(const Stats& stats) override;

  void SetHasVideo() override { hasVideo_ = true; }
  void SetHasAudio() override { hasAudio_ = true; }
  bool HasVideo() override { return hasVideo_; }
  bool HasAudio() override { return hasAudio_; }
  bool HasVideoInformed() override { return hasVideoInformed; }
  bool HasAudioInformed() override { return hasAudioInformed; }
 private:
  friend class LEBConnection;
  explicit LEBConnectionImpl(LEBConnection::Client* client);

  class PeerConnectionDelegateHandler : public RTCPeerConnectionDelegate {
   public:
    explicit PeerConnectionDelegateHandler(LEBConnectionImpl* pc_impl);
    ~PeerConnectionDelegateHandler();

    // Implement RTCPeerConnectionDelegate
    void onChangeConnectionState(RTCPeerConnection* pc,
                                 RTCPeerConnectionState state) override;

    void onAddReceiver(RTCPeerConnection* pc,
                       std::shared_ptr<RTCRtpReceiver> receiver) override;

    void onChangeCandidate(RTCPeerConnection* pc, RTCIceCandidate& local,
                           RTCIceCandidate& remote, int last_data_received_ms,
                           std::string reason) override;

    void onStatsReport(RTCStatsReport& stats) override;

    void onError(RTCPeerConnection* pc, const RTCError* error) override;

   private:
    LEBConnectionImpl* pc_impl_;
  };

  LEBConnection::Client* client_;

  std::shared_ptr<PeerConnectionDelegateHandler> pc_delegate_hander_;

  std::shared_ptr<RTCPeerConnection> pc_;

  std::shared_ptr<SessionDescriptionExchange> sdp_exchange_;

  std::map<RTCRtpMediaType, std::shared_ptr<ChannelReceiver>> receivers_;

  uint64_t start_time_;
  std::string data_server_ip_;
  std::string signal_server_ip_;
  int64_t signal_dns_cost_ms_;
  int64_t first_video_received_time_;
  int64_t first_audio_received_time_;

  bool hasVideo_ = false;
  bool hasAudio_ = false;
  bool hasVideoInformed = false;
  bool hasAudioInformed = false;
};

}  // namespace lebnetwork
}  // namespace liteav

#endif  // LEB_NETWORK_LEB_CONNECTION_IMPL_H_
