// Copyright (c) 2021 Tencent. All rights reserved.

#ifndef LEB_NETWORK_LEB_DEFINES_H_
#define LEB_NETWORK_LEB_DEFINES_H_

#include <string>

#include "include/rtc_error.h"
#include "leb_network/base/copy_on_write_buffer.h"

namespace liteav {
namespace lebnetwork {

using RTCError = media_transport::RTCError;
using RTCErrorType = media_transport::RTCErrorType;
using RTCErrorCode = media_transport::RTCErrorCode;

enum class VideoCodecType {
  kNoVideo = -1,
  kH264,
  kH265,
};

enum class AudioCodecType {
  kNoAudio = -1,
  kOpus,
  kAac,
};

enum class AudioFrameType {
  kNone = -1,
  kOpus,
  kAacAdts,
  kAacAdifHeader,
  kAacAdifData,
};

enum class MetadataType {
  kAudio,
  kVideo,
  kData,
};

enum class NetState {
  // 正在连接中
  kConnecting,
  // 连接成功
  kConnected,
  // 当前已经断网，正在尝试重新连接
  kDisconnected,
  // 重连失败，网络不可恢复
  kConnectFailed,
  // 客户主动关闭连接
  kConnectClosed,
};

enum class LogLevel {
  kDebug,
  kInfo,
  kWarning,
  kError,
};

enum class SdpType {
  kOffer,
  kAnswer,
};

struct SdpInfo {
  SdpInfo();
  SdpInfo(SdpInfo&&);
  ~SdpInfo();
  SdpInfo& operator=(SdpInfo&&);

  SdpType sdp_type;
  std::string sdp;
};

struct VideoInfo {
  VideoInfo();
  VideoInfo(VideoInfo&&);
  ~VideoInfo();
  VideoInfo& operator=(VideoInfo&&);

  VideoCodecType codec_type = VideoCodecType::kH264;
  uint8_t extra_data[4096];
  size_t extra_size = 0;
};

struct AudioInfo {
  AudioInfo();
  AudioInfo(AudioInfo&&);
  ~AudioInfo();
  AudioInfo& operator=(AudioInfo&&);

  AudioCodecType codec_type = AudioCodecType::kAac;
  int32_t sample_rate = 0;
  int32_t num_channels = 0;
};

struct EncodedVideoFrame {
  EncodedVideoFrame();
  EncodedVideoFrame(EncodedVideoFrame&&);
  ~EncodedVideoFrame();
  EncodedVideoFrame& operator=(EncodedVideoFrame&&);

  trtc::CopyOnWriteBuffer data;
  int64_t pts = 0;
  int64_t dts = 0;
  VideoCodecType codec_type = VideoCodecType::kNoVideo;
  bool is_key_frame = false;
  uint64_t sequence_number = 0;
  int32_t output_delay_ms = 0;
};

struct EncodedAudioFrame {
  EncodedAudioFrame();
  EncodedAudioFrame(EncodedAudioFrame&&);
  ~EncodedAudioFrame();
  EncodedAudioFrame& operator=(EncodedAudioFrame&&);

  trtc::CopyOnWriteBuffer data;
  int64_t pts = 0;
  int32_t sample_rate = 0;
  int32_t num_channels = 0;
  int32_t bits_per_sample = 0;
  AudioCodecType codec_type = AudioCodecType::kNoAudio;
  AudioFrameType frame_type = AudioFrameType::kNone;
  int32_t output_delay_ms = 0;
};

struct Metadata {
  Metadata();
  Metadata(Metadata&&);
  ~Metadata();
  Metadata& operator=(Metadata&&);

  trtc::CopyOnWriteBuffer data;
  MetadataType type = MetadataType::kData;
};

struct Stats {
  // network
  int32_t rtt_ms = 0;
  // 连接耗时
  // int32_t connect_cost_ms = 0;
  // 信令服务器 IP
  std::string signal_server_ip;
  // 信令服务器域名解析耗时
  int32_t signal_dns_cost_ms = 0;
  // 数据服务器 IP
  std::string data_server_ip;
  // 从开始拉流到收到首个视频帧的时间
  int32_t first_video_received_cost_ms = 0;
  // 从开始拉流到收到首个音频帧的时间
  int32_t first_audio_received_cost_ms = 0;

  // video
  uint32_t video_packets_received = 0;
  uint64_t video_bytes_received = 0;
  int32_t video_packets_lost = 0;
  uint32_t video_nack_count = 0;

  // audio
  uint32_t audio_packets_received = 0;
  uint64_t audio_bytes_received = 0;
  int32_t audio_packets_lost = 0;
  uint32_t audio_nack_count = 0;
};

class Error : public RTCError {};

}  // namespace lebnetwork
}  // namespace liteav

#endif  // LEB_NETWORK_LEB_DEFINES_H_
