// Copyright (c) 2021 Tencent. All rights reserved.

#include "leb_network/session_description_exchange.h"

#include "include/peer_connection.h"
#include "leb_network/internal_defines.h"

namespace {

std::string SdpTypeToString(media_transport::RTCSdpType type) {
  switch (type) {
    case media_transport::RTCSdpType::RTCSdpTypeOffer:
      return "local_sdp";
    case media_transport::RTCSdpType::RTCSdpTypePrAnswer:
      return "server_remote_sdp";
    case media_transport::RTCSdpType::RTCSdpTypeAnswer:
      return "final_remote_sdp";
  }
}

}  // namespace

namespace liteav {
namespace lebnetwork {

SessionDescriptionExchange::SessionDescriptionExchange(
    std::shared_ptr<RTCPeerConnection> pc, SDPExchangeObserver* sdp_observer,
    LEBConnection::Client* error_observer)
    : pc_(pc), error_observer_(error_observer), sdp_observer_(sdp_observer) {}

SessionDescriptionExchange::~SessionDescriptionExchange() = default;

void SessionDescriptionExchange::CreateOffer() {
  pc_->setSessionDescriptionDelegate(shared_from_this());
  pc_->createOffer();
}

void SessionDescriptionExchange::SetRemoteSDP(RTCSessionDescription& sdp) {
  remote_description_ = sdp;
  pc_->startPeerConnection(&local_description_, &remote_description_);
}

void SessionDescriptionExchange::NotifyError(const RTCError* error) {
  Error e = RTCErrorToError(*error);
  error_observer_->OnError(e);
}

void SessionDescriptionExchange::onCreateSessionDescription(
    RTCPeerConnection* pc, const RTCSessionDescription* sdp, const RTCError* error,
    void* user_data) {
  if (pc_.get() != pc) {
    return;
  }

  if (error->ok()) {
    NotifyLog("", LogLevel::kInfo, "SDP type:%s, msg:%s",
              SdpTypeToString(sdp->type).c_str(), sdp->sdp.c_str());
  } else {
    NotifyLog("", LogLevel::kError, "SDP exchange error. type:%s, msg:%s",
              SdpTypeToString(sdp->type).c_str(), error->message());

    NotifyError(error);
  }

  if (sdp->type == RTCSdpType::RTCSdpTypeOffer) {
    local_description_ = *sdp;
    sdp_observer_->OnOfferCreated(local_description_);
  } else if (sdp->type == RTCSdpType::RTCSdpTypeAnswer) {
    remote_description_ = *sdp;
  }
}

void SessionDescriptionExchange::onSignalingResolve(
    RTCPeerConnection* pc, const std::string& server_ip,
    const int64_t dns_cost_ms) {
  if (pc_.get() != pc) {
    return;
  }

  sdp_observer_->OnSignaling(server_ip, dns_cost_ms);
}

void SessionDescriptionExchange::onCompletion(RTCPeerConnection* pc,
                                              const RTCError* error,
                                              void* user_data) {
  if (pc_.get() != pc) {
    return;
  }
  if (error->ok()) {
    NotifyLog("", LogLevel::kInfo, "SDP exchange completely.");
  } else {
    NotifyLog("", LogLevel::kError, "SDP exchange error. msg:%s", error->message());

    NotifyError(error);
  }
}

}  // namespace lebnetwork
}  // namespace liteav
