// Copyright (c) 2021 Tencent. All rights reserved.

#include "leb_network/video_channel_receiver.h"

#include "leb_network/converter/h264_converter.h"
#include "leb_network/converter/h265_converter.h"

namespace liteav {
namespace lebnetwork {

VideoChannelReceiver::VideoChannelReceiver(
    std::shared_ptr<RTCPeerConnection> pc, LEBConnection::Client* client)
    : ChannelReceiver(pc, client), throtter_(60000), hasInformed_(false) {}

VideoChannelReceiver::~VideoChannelReceiver() = default;

// Implement RTCRtpReceiverDelegate
void VideoChannelReceiver::OnReceiveFirstPacket(RTCRtpMediaType mediaType) {}

int VideoChannelReceiver::getEndSeqNumOfLastDecodedFrame(
    RTCRtpMediaType mediaType, uint16_t* sequence_number, uint32_t* timestamp) {
  return -1;
}

void VideoChannelReceiver::OnReceiveAudioCodecs(
    const std::map<int, SdpAudioFormat>& codecs) {}

void VideoChannelReceiver::OnReceiveVideoCodecs(
    const std::map<int, SdpVideoFormat>& codecs) {
  formats_ = codecs;
  client_->SetHasVideo();

#if 0
  // outband update sps pps
  auto it = codecs.begin();
  while (it != codecs.end() && !hasInformed_) {
    VideoInfo info;
    if (it->second.codecType == media_transport::VideoCodecType::kVideoCodecH264) {

      SdpVideoFormat format = it->second;
      auto iter = format.parameters.find("sprop-parameter-sets");
      if (iter != format.parameters.end()) {

      }
      // info.codec_type = VideoCodecType::kH264;
      // client_->OnVideoInfo(info);
      // hasInformed_ = true;
    } else if (it->second.codecType == media_transport::VideoCodecType::kVideoCodecH265) {
      SdpVideoFormat format = it->second;
      auto iter = format.parameters.find("sprop-vps");
      if (iter != format.parameters.end()) {

      }
      auto iter = format.parameters.find("sprop-sps");
      if (iter != format.parameters.end()) {

      }
      auto iter = format.parameters.find("sprop-pps");
      if (iter != format.parameters.end()) {

      }
      // info.codec_type = VideoCodecType::kH265;
      // client_->OnVideoInfo(info);
      // hasInformed_ = true;
    } else {
      NotifyLog("", LogLevel::kWarning, "Unknown video codec %d", it->second.codecType);
    }

    it++;
  }
#endif
}

void VideoChannelReceiver::OnReceiveVideoSpropParameterSets(const uint8_t* extradata, size_t size) {
  // inband update sps pps
  auto it = formats_.begin();
  while (it != formats_.end() && !hasInformed_) {
    VideoInfo info;
    if (size > 0) {
      memcpy(info.extra_data, extradata, size);
      info.extra_size = size;
    } else {
      info.extra_size = 0;
    }
    if (it->second.codecType == media_transport::VideoCodecType::kVideoCodecH264) {
      info.codec_type = VideoCodecType::kH264;
      client_->OnVideoInfo(info);
      hasInformed_ = true;
    } else if (it->second.codecType == media_transport::VideoCodecType::kVideoCodecH265) {
      info.codec_type = VideoCodecType::kH265;
      client_->OnVideoInfo(info);
      hasInformed_ = true;
    } else {
      NotifyLog("", LogLevel::kWarning, "Unknown video codec %d", it->second.codecType);
    }

    it++;
  }
}

// Implement RtpPacketSinkInterface
void VideoChannelReceiver::OnRtpPacket(const RtpPacket& packet,
                                       RTCRtpMediaType mediaType) {
  EnsureConverterGenerated(packet.header.payloadType);
  if (converter_) {
    converter_->Append(packet);

    if (client_->HasVideo() && !client_->HasVideoInformed())
      return;

    std::vector<std::unique_ptr<EncodedVideoFrame>> frames =
        converter_->Convert();
    for (auto& frame : frames) {
      if (!hasInformed_) {
        VideoInfo info;
        info.codec_type = frame->codec_type;
        client_->OnVideoInfo(info);;
        hasInformed_ = true;
      }
      client_->OnEncodedVideo(std::move(frame));
    }
  }
}

void VideoChannelReceiver::OnRtpMetaDataPacket(const RtpMetaData& packet,
                                               RTCRtpMediaType mediaType) {
  std::unique_ptr<Metadata> data(new Metadata());
  switch (mediaType) {
    case RTCRtpMediaType::RTCRtpMediaTypeAudio:
      data->type = MetadataType::kAudio;
      break;
    case RTCRtpMediaType::RTCRtpMediaTypeVideo:
      data->type = MetadataType::kVideo;
      break;
#ifdef WEBRTC_BUILD_DATA_CHANNEL
    case RTCRtpMediaType::RTCRtpMediaTypeData:
      data->type = MetadataType::kData;
      break;
#endif
  }

  auto md = packet.getByteMetaData();
  data->data.SetData(&(md[0]), md.size());

  client_->OnMetadata(std::move(data));
}

void VideoChannelReceiver::OnMediaPacket(const void* data, int length,
                                         int64_t pts, RTCRtpMediaType mediaType,
                                         bool iscodec) {}

void VideoChannelReceiver::EnsureConverterGenerated(uint8_t payload_type) {
  if (converter_) {
    return;
  }

  auto iter = formats_.find(payload_type);
  if (iter != formats_.end()) {
    if (iter->second.codecType ==
        media_transport::VideoCodecType::kVideoCodecH264) {
      converter_.reset(new H264Converter());
    } else if (iter->second.codecType ==
               media_transport::VideoCodecType::kVideoCodecH265) {
      converter_.reset(new H265Converter(iter->second));
    } else {
      if (throtter_.ShouldGo()) {
        NotifyLog("", LogLevel::kWarning, "Unknown video type %d",
                  static_cast<int>(iter->second.codecType));
      }
    }
  } else {
    if (throtter_.ShouldGo()) {
      NotifyLog("", LogLevel::kWarning, "Unknown video payload type %d",
                static_cast<int>(payload_type));
    }
  }
}

}  // namespace lebnetwork
}  // namespace liteav
