// Copyright (c) 2020 Tencent. All rights reserved.

#include "leb_connection_test.h"
#ifdef WEBRTC_ANDROID
#include <android/log.h>
#endif

namespace {

using LogLevel = liteav::lebnetwork::LogLevel;
std::string LogLevelToString(LogLevel level) {
  switch (level) {
    case LogLevel::kDebug:
      return "[Debug]";
    case LogLevel::kInfo:
      return "[Info]";
    case LogLevel::kWarning:
      return "[Warning]";
    case LogLevel::kError:
      return "[Error]";
  }
}
}  // namespace

namespace liteav {
namespace lebnetwork {

LEBConnectionTest::LEBConnectionTest(const std::string& store_dir)
    : store_dir_(store_dir), audio_file_(nullptr), video_file_(nullptr) {
  LEBConnection::Initialize(this, LogLevel::kInfo);
}

LEBConnectionTest::~LEBConnectionTest() { LEBConnection::Uninitialize(); }

void LEBConnectionTest::Start(const LEBConnection::Configuration& config) {
  if (pc_) {
    return;
  }
  pc_ = LEBConnection::Create(this);
  pc_->Start(config);

  std::string audio_filename = store_dir_ + "audio.aac";
  audio_file_ = fopen(audio_filename.c_str(), "wb");
  std::string video_filename = store_dir_ + "video.h264";
  video_file_ = fopen(video_filename.c_str(), "wb");
}

void LEBConnectionTest::Stop() {
  if (!pc_) {
    return;
  }
  pc_->Stop();
  pc_.reset();

  fclose(audio_file_);
  fclose(video_file_);
  audio_file_ = nullptr;
  video_file_ = nullptr;
}

void LEBConnectionTest::GetStats() {
  if (pc_) {
    pc_->GetStats();
  }
}

// Implement LEBConnection::Client
void LEBConnectionTest::OnEncodedVideo(
    std::unique_ptr<EncodedVideoFrame> frame) {
  fwrite(frame->data.cdata(), 1, frame->data.size(), video_file_);
  fflush(video_file_);
  printf(
      "LEBConnectionTest Video data pts:%" PRId64" dts:%" PRId64" type:%d is_key:%d "
      "size:%zu\n",
      frame->pts, frame->dts, static_cast<int>(frame->codec_type),
      static_cast<int>(frame->is_key_frame), frame->data.size());
}

void LEBConnectionTest::OnEncodedAudio(
    std::unique_ptr<EncodedAudioFrame> frame) {
  fwrite(frame->data.cdata(), 1, frame->data.size(), audio_file_);
  fflush(audio_file_);
  printf(
      "LEBConnectionTest Audio data pts:%" PRId64" samples:%d channels:%d codec:%d "
      "type:%d size:%zu\n",
      frame->pts, frame->sample_rate, frame->num_channels,
      static_cast<int>(frame->codec_type), static_cast<int>(frame->frame_type),
      frame->data.size());
}

void LEBConnectionTest::OnMetadata(std::unique_ptr<Metadata> data) {
  printf("LEBConnectionTest Metadata size:%zu\n", data->data.size());
}

void LEBConnectionTest::OnError(const Error& error) {
  printf("LEBConnectionTest OnError: %s\n", error.message());
}

void LEBConnectionTest::OnNetState(NetState state) {
  printf("LEBConnectionTest OnNetState %d\n", static_cast<int>(state));
}

void LEBConnectionTest::OnStats(const Stats& stats) {
  printf(
      "LEBConnectionTest OnStats rtt:%d audio_bytes:%" PRIu64" audio_nack:%u "
      "audio_lost:%d "
      "audio_pkts:%u video_bytes:%" PRIu64" video_nack:%u video_lost:%d "
      "video_pkts:%u\n",
      stats.rtt_ms, stats.audio_bytes_received, stats.audio_nack_count,
      stats.audio_packets_lost, stats.audio_packets_received,
      stats.video_bytes_received, stats.video_nack_count,
      stats.video_packets_lost, stats.video_packets_received);
}

// Implement LEBConnection::LogObserver
void LEBConnectionTest::OnLog(const char* tag, LogLevel level, const char* msg) {
#ifdef WEBRTC_ANDROID
  std::string TAG("[LEBConnectionTest]");
  switch(level) {
    case LogLevel::kDebug:
      __android_log_print(ANDROID_LOG_VERBOSE, TAG.c_str(), "%s", msg);
      break;
    case LogLevel::kInfo:
      __android_log_print(ANDROID_LOG_INFO, TAG.c_str(), "%s", msg);
      break;
    case LogLevel::kWarning:
      __android_log_print(ANDROID_LOG_WARN, TAG.c_str(), "%s", msg);
      break;
    case LogLevel::kError:
      __android_log_print(ANDROID_LOG_ERROR, TAG.c_str(), "%s", msg);
      break;
  }
#else 
  printf("LEBConnectionTest Log: %s %s\n", LogLevelToString(level).c_str(), msg);
#endif
}

}  // namespace lebnetwork
}  // namespace liteav
