// Copyright (c) 2021 Tencent. All rights reserved.

#ifndef PEER_CONNECTION_H
#define PEER_CONNECTION_H

#include <string>

#include "rtc_error.h"
#include "rtc_receiver.h"
#include "rtc_configuration.h"
#include "rtc_stats_report.h"
#include "rtc_macros.h"
#include <functional>
#include <mutex>

namespace media_transport {

class RTCRtpReceiver;
class RTCStatisticsReport;
class RTCLegacyStatsReport;

/** Represents the signaling state of the peer connection. */
typedef enum RTCSignalingState {
  RTCSignalingStateStable,
  RTCSignalingStateHaveLocalOffer,
  RTCSignalingStateHaveLocalPrAnswer,
  RTCSignalingStateHaveRemoteOffer,
  RTCSignalingStateHaveRemotePrAnswer,
  // Not an actual state, represents the total number of states.
  RTCSignalingStateClosed,
} RTCSignalingState;

/** Represents the ice connection state of the peer connection. */
typedef enum RTCIceConnectionState {
  RTCIceConnectionStateNew,
  RTCIceConnectionStateChecking,
  RTCIceConnectionStateConnected,
  RTCIceConnectionStateCompleted,
  RTCIceConnectionStateFailed,
  RTCIceConnectionStateDisconnected,
  RTCIceConnectionStateClosed,
  RTCIceConnectionStateCount,
} RTCIceConnectionState;

/** Represents the combined ice+dtls connection state of the peer connection. */
typedef enum RTCPeerConnectionState {
  RTCPeerConnectionStateNew,
  RTCPeerConnectionStateConnecting,
  RTCPeerConnectionStateConnected,
  RTCPeerConnectionStateDisconnected,
  RTCPeerConnectionStateFailed,
  RTCPeerConnectionStateClosed,
} RTCPeerConnectionState;

/** Represents the ice gathering state of the peer connection. */
typedef enum RTCIceGatheringState {
  RTCIceGatheringStateNew,
  RTCIceGatheringStateGathering,
  RTCIceGatheringStateComplete,
} RTCIceGatheringState;

/** Represents the stats output level. */
typedef enum RTCStatsOutputLevel {
  RTCStatsOutputLevelStandard,
  RTCStatsOutputLevelDebug,
} RTCStatsOutputLevel;

typedef enum RTCSdpType {
  RTCSdpTypeOffer,
  RTCSdpTypePrAnswer,
  RTCSdpTypeAnswer,
} RTCSdpType;

class RTCPeerConnection;

struct RTCSessionDescription {
  RTCSdpType type;

  std::string sdp;
};


struct RTCIceCandidate {
  /**
  * If present, the identifier of the "media stream identification" for the media
  * component this candidate is associated with.
  */
  std::string sdpMid;
  
  /**
  * The index (starting at zero) of the media description this candidate is
  * associated with in the SDP.
  */
  int sdpMLineIndex;
  
  /** The SDP string for this candidate. */
  std::string sdp;
  
  /** The URL of the ICE server which this candidate is gathered from. */
  std::string serverUrl;
  
  /** The IP of the ICE server which this candidate is gathered from. */
  std::string serverIp;
};

/**
 * PeerConnection回调接口 
 */
class RTC_EXPORT_API RTCPeerConnectionDelegate {
public:
  /** Called when the SignalingState changed. */
  virtual void onChangeSignalingState(RTCPeerConnection * peerConnection,
                                      RTCSignalingState stateChanged){}
  
  /** Called any time the IceConnectionState changes. */
  virtual void onChangeIceConnectionState(RTCPeerConnection * peerConnection,
                                          RTCIceConnectionState stateChanged){}
  
  /** Called any time the IceGatheringState changes. */
  virtual void onChangeIceGatheringState(RTCPeerConnection * peerConnection,
                                         RTCIceGatheringState stateChanged){}
  
  /** Called any time the PeerConnectionState changes. */
  virtual void onChangeConnectionState(RTCPeerConnection * peerConnection,
                                       RTCPeerConnectionState stateChanged){}
  
  /** New ice candidate has been found. */
  virtual void onGenerateIceCandidate(RTCPeerConnection *peerConnection,
                                      RTCIceCandidate &candidate) {}

  /** Called when a group of local Ice candidates have been removed. */
  virtual void onRemoveIceCandidates(RTCPeerConnection *peerConnection, std::vector<RTCIceCandidate> &candidates){}
  
  /** Called when the selected ICE candidate pair is changed. */
  virtual void onChangeCandidate(RTCPeerConnection *peerConnection, RTCIceCandidate &local, RTCIceCandidate &remote,
                              int lastDataReceivedMs, std::string reason) {}
  
  /** Called when a receiver and its media streams are created. */
  virtual void onAddReceiver(RTCPeerConnection * peerConnection,
                             std::shared_ptr<RTCRtpReceiver> receiver) = 0;
  
  /** Called when getStats() required */
  virtual void onStatsReport(RTCStatsReport &stats) = 0;
  
  /** Called when error happend */
  virtual void onError(RTCPeerConnection * peerConnection,
                       const RTCError *error) = 0;
  
  virtual ~RTCPeerConnectionDelegate() = default;
};


class RTC_EXPORT_API RTCSessionDescriptionDelegate {
public:
  virtual void onCreateSessionDescription(RTCPeerConnection *peerConnection, const RTCSessionDescription *sdp,
                                          const RTCError *error, void *userData) = 0;
  
  virtual void onCompletion(RTCPeerConnection *peerConnection, const RTCError *error, void *userData) = 0;
  
  virtual void onSignalingResolve(RTCPeerConnection * peerConnection,
                                const std::string &streamIp, const int64_t resolveCostMs){}
  
  virtual ~RTCSessionDescriptionDelegate() = default;
};

typedef std::function< void(RTCSessionDescription *sdp, const RTCError *error, void *userData)> CreateSessionDescriptionHandlerFun;
typedef std::function<void(const RTCError *error, void *userData)> CompletionHandlerFun;
typedef std::function< void(RTCSessionDescription *sdp, const RTCError *error)> CreateSessionDescriptionHandler;
typedef std::function<void(const RTCError *error)> CompletionHandler;

class RTC_EXPORT_API RTCPeerConnection {
public:
  RTCPeerConnection() = default;
  virtual ~RTCPeerConnection() = default;
  
  /** 设置配置参数 */
  void setConfiguration(RTCConfiguration &configure);

  RTCConfiguration getConfiguration();
  
  /** 创建offer并传入自定义回调处理函数，offer sdp回调输出，在自定义回调处理函数中可以开始信令请求 */
  void createOffer() {
      createOfferInternal();
  }

  /** 开始peer connection连接 */
  void startPeerConnection(RTCSessionDescription *localSDP,
                           RTCSessionDescription *remoteSDP) {
    startPeerConnectionInternal(localSDP, remoteSDP);
  }

  /** 停止peer connection连接 */
  void stopPeerConnection();
  
  //TO-DO
  bool isRunning();
  
  /** 释放peer connection相关资源 */
  void closePeerConnection();
  
  /** 设置RTCPeerConnectionDelegate */
  void setDelegate(std::weak_ptr<RTCPeerConnectionDelegate> delegate);
  
  void setSessionDescriptionDelegate(std::weak_ptr<RTCSessionDescriptionDelegate> sessionDelegate);
  
  
  std::weak_ptr<RTCSessionDescriptionDelegate> getSessionDescriptionDelegate();
  /** 请求统计信息，onStatsReport回调输出 */
  void getStats();
  
  void setUserData(void *userdata);
  
  void* getUserData();
  
private:
  void createOfferInternal();
  void startPeerConnectionInternal(RTCSessionDescription *localSDP, RTCSessionDescription *sdp);
  
  RTCPeerConnection(const RTCPeerConnection&) = delete;
  RTCPeerConnection& operator=(const RTCPeerConnection&) = delete;
  
  std::shared_ptr<class PeerConnectionAdapter> _adapter;
  
  std::weak_ptr<RTCSessionDescriptionDelegate> _sessionDelegate;
  
  void *_userdata;
  
  RTCConfiguration _config;
  
  std::mutex _dataLock;
  std::mutex _adapterLock;
};

}
#endif /* PEER_CONNECTION_H */
