// Copyright (c) 2021 Tencent. All rights reserved.

#ifndef RTC_CONFIGURATION_H
#define RTC_CONFIGURATION_H

#include <inttypes.h>
#include <stdio.h>

#include <functional>
#include <map>
#include <mutex>
#include <vector>

#include "rtc_macros.h"

namespace media_transport {

typedef enum {
    MediaTransportLogLevelDebug,
    MediaTransportLogLevelInfo,
    MediaTransportLogLevelWarning,
    MediaTransportLogLevelError,
    MediaTransportLogLevelNone,
} MediaTransportLogLevel;

void MediaTransportLogEx(MediaTransportLogLevel severity, const char* log_string);

#define MediaTransportLogString(buf, format, ...)                       \
  snprintf(buf, 1024, "(%s:%d %s): " format, __FILE__,                  \
                __LINE__, __FUNCTION__, ##__VA_ARGS__);

#define MediaTransportLogFormat(severity, format, ...)                     \
  do {                                                          \
    char log_string[1024] = {0};   \
    MediaTransportLogString(log_string, format, ##__VA_ARGS__); \
    MediaTransportLogEx(severity, log_string);                             \
  } while (false)

#define MediaTransportLogDebug(format, ...) \
  MediaTransportLogFormat(media_transport::MediaTransportLogLevel::MediaTransportLogLevelDebug, format, ##__VA_ARGS__)

#define MediaTransportLogInfo(format, ...) \
  MediaTransportLogFormat(media_transport::MediaTransportLogLevel::MediaTransportLogLevelInfo, format, ##__VA_ARGS__)

#define MediaTransportLogWarning(format, ...) \
  MediaTransportLogFormat(media_transport::MediaTransportLogLevel::MediaTransportLogLevelWarning, format, ##__VA_ARGS__)

#define MediaTransportLogError(format, ...) \
  MediaTransportLogFormat(media_transport::MediaTransportLogLevel::MediaTransportLogLevelError, format, ##__VA_ARGS__)

#define MediaTransportLog(format, ...) MediaTransportLogInfo(format, ##__VA_ARGS__)

typedef enum {
  kNetUnknown   = 0,
  kNetEthernet  = 1 << 0,   // 以太网（有线）
  kNetWifi      = 1 << 1,
  kNetCellular  = 1 << 2,   // 移动网络（4G）
  kNetVpn       = 1 << 3,
  kNetLoopback  = 1 << 4,
} NetworkType;

class RTC_EXPORT_API RTCConfiguration {
public:
  RTCConfiguration() = default;
  RTCConfiguration(const RTCConfiguration& other) = default;

  // 视频开关，true表示拉取视频流
  bool receiveVideo = true;

  // 音频开关，true表示拉取音频流
  bool receiveAudio = true;
  
  // 加密开关
  bool disableEncryption = false;

  // 是否开启AAC选项，打开且推流音频是AAC，CDN下发AAC，否则默认音频是Opus
  bool enableAAC = false;

  // 是否开启flexfec
  bool enableFlexFec = false;

  // 是否开启音频补包
  bool enableAudioPlc = false;

  // 是否开启0RTT数据下发
  bool enable0Rtt = false;

  // 音频包排序最大延时，一般300~1000ms
  int audioMaxReorderDelayMs = 500;

  // 是否开启UDP信令，如开启需要设置下面地址和端口
  bool enableUdpSignal = false;
  std::string remoteAddress = "webrtc.liveplay.myqcloud.com"; //远程ip或域名，推荐用ip
  int remotePort = 8000; //固定远程端口
  // 默认远程监听端口为：8000, 14000，如后台有特殊配置可以通过此接口增加其他端口
  void AddExtraRemotePorts(std::vector<int> ports);
  std::vector<int> GetExtraRemotePorts();

  std::string streamUrl;

  // 统计信息回调周期
  int stats_period_ms = 5000;

  // 拉流超时，answer或第一份数据超时
  int start_timeout_ms = 5000;

  // 首个GOP有回退时下发的数据量，需要后台配置并在CDN命中缓存时生效
  int min_delay_ms = 1000;

  // 网络类型，后台可以据此进行不同的发包策略
  NetworkType network_type = kNetUnknown;

  // only used for android
  void AddNetworkHandle(std::string ip, int64_t handle);
  int64_t GetNetworkHandle(std::string ip);

  std::string ToString();

private:
  std::map<std::string, int64_t> network_handle_;
  std::vector<int> extraRemotePorts_;
};

typedef std::function<void(const char*, MediaTransportLogLevel, const char*)> RTCLogCallbackFunc;

static const int64_t  kDefaultMaxFileSize = 10 * 1024 * 1024; // 10MB.

class RTC_EXPORT_API RTCFileLogger {
public:
  //deprecated
  static RTCFileLogger* create(const char* dirPath, uint32_t maxFileSize = kDefaultMaxFileSize);
  static std::shared_ptr<RTCFileLogger> getInstance(const char* dirPath, uint32_t maxFileSize = kDefaultMaxFileSize);

protected:
  RTCFileLogger() = default;
  virtual ~RTCFileLogger()=default;
  
public:
  class RTCLogInterface {
  public:
    RTCLogInterface() = default;
    virtual ~RTCLogInterface()=default;
    
    virtual void OnLog(const char* tag,
                       MediaTransportLogLevel level,
                       const char* msg) = 0;
  };
  
  virtual void setDisableBuffering(bool flag) = 0;
  
  virtual void setLogLevel(MediaTransportLogLevel loglevel) = 0;

  virtual void setLogCallback(RTCLogCallbackFunc callback) = 0;
  
  virtual void setLogInterface(std::weak_ptr<RTCLogInterface> callback) = 0;
  
  virtual void start() = 0;

  // Stops writing WebRTC logs to disk. This method is also called on dealloc.
  //deprecated
  virtual void stop() = 0;
private:
  RTCFileLogger& operator=(const RTCFileLogger&) = delete;
  RTCFileLogger(const RTCFileLogger& singleton2) = delete;
  
  static std::shared_ptr<RTCFileLogger> sinstance;
};


class RTCUtils {
public:
  static int64_t TimeMillis();

  static int64_t TimeMicros();

  static int64_t TimeNanos();
  
protected:
  RTCUtils() = default;
  virtual ~RTCUtils()=default;
  
};

#define MediaTransportFreqLog(severity, maxcost, format, ...) do {  \
                                                        \
                                                        \
static int count = 0;                                   \
static int64_t begin = 0;                               \
int64_t cost = -1;                                       \
                                                        \
if (begin == 0) {                                       \
  begin = RTCUtils::TimeMillis();                       \
} else {                                                \
  cost = RTCUtils::TimeMillis() - begin;                \
}                                                       \
if (cost == -1 || cost > maxcost) {                     \
 char log_string[1024] = {0};                            \
 snprintf(log_string, 1024, "(%s cost:%" PRId64":count:%d): " format, __FUNCTION__, cost == -1 ? 0 : cost, count, ##__VA_ARGS__);                              \
 MediaTransportLogEx(severity, log_string);             \
                                                        \
if (cost != -1) {begin = 0;count=0;}                    \
                                                        \
} else {                                                \
  count++;                                              \
}                                                       \
                                                        \
} while(0)                                              \

#define MediaTransportFreqLogWarning(format, ...) MediaTransportFreqLog(media_transport::MediaTransportLogLevel::MediaTransportLogLevelWarning, 10000, format, ##__VA_ARGS__)
#define MediaTransportFreqLogInfo(format, ...) MediaTransportFreqLog(media_transport::MediaTransportLogLevel::MediaTransportLogLevelInfo, 10000, format, ##__VA_ARGS__)

}
#endif /* RTC_CONFIGURATION_H */
