// Copyright (c) 2021 Tencent. All rights reserved.

#ifndef RTC_ERROR_H
#define RTC_ERROR_H

#include <string>
#include "rtc_macros.h"

namespace media_transport {

enum class RTCErrorType {
  // No error.
  NONE,

  // An operation is valid, but currently unsupported.
  // Maps to OperationError DOMException.
  UNSUPPORTED_OPERATION,

  // A supplied parameter is valid, but currently unsupported.
  // Maps to OperationError DOMException.
  UNSUPPORTED_PARAMETER,

  // General error indicating that a supplied parameter is invalid.
  // Maps to InvalidAccessError or TypeError DOMException depending on context.
  INVALID_PARAMETER,

  // Slightly more specific than INVALID_PARAMETER; a parameter's value was
  // outside the allowed range.
  // Maps to RangeError DOMException.
  INVALID_RANGE,

  // Slightly more specific than INVALID_PARAMETER; an error occurred while
  // parsing string input.
  // Maps to SyntaxError DOMException.
  SYNTAX_ERROR,

  // The object does not support this operation in its current state.
  // Maps to InvalidStateError DOMException.
  INVALID_STATE,

  // An attempt was made to modify the object in an invalid way.
  // Maps to InvalidModificationError DOMException.
  INVALID_MODIFICATION,

  // An error occurred within an underlying network protocol.
  // Maps to NetworkError DOMException.
  NETWORK_ERROR,

  // Some resource has been exhausted; file handles, hardware resources, ports,
  // etc.
  // Maps to OperationError DOMException.
  RESOURCE_EXHAUSTED,

  // The operation failed due to an internal error.
  // Maps to OperationError DOMException.
  INTERNAL_ERROR,

  MEDIA_TRANSPORT_EXTEND_ERROR = 1000
};

enum class RTCErrorCode {
  // No error
  NONE,
  // DNS解析出错
  RESOLVE_ERROR,
  // stream url格式出错
  SCHEME_ERROR,
  // 鉴权失败
  AUTH_ERROR,
  // 码流不存在
  NOTFOUND_ERROR,
  // 未知错误
  UNKNOW_ERROR,
  // 信令超时
  SIGNAL_TIMEOUT,
  // 数据超时
  DATA_TIMEOUT,
};

class RTC_EXPORT_API RTCError {
 public:
  // Constructors.

  // Creates a "no error" error.
  RTCError() {}
  explicit RTCError(RTCErrorType type) : type_(type) {}

  RTCError(RTCErrorType type, std::string message)
      : type_(type), message_(std::move(message)) {}

  // Delete the copy constructor and assignment operator; there aren't any use
  // cases where you should need to copy an RTCError, as opposed to moving it.
  // Can revisit this decision if use cases arise in the future.
  RTCError(const RTCError& other) = delete;
  RTCError& operator=(const RTCError& other) = delete;

  // Move constructor and move-assignment operator.
  RTCError(RTCError&& other);
  RTCError& operator=(RTCError&& other);

  // Identical to default constructed error.
  //
  // Preferred over the default constructor for code readability.
  static RTCError OK();

  // Error type.
  RTCErrorType type() const { return type_; }
  void set_type(RTCErrorType type) { type_ = type; }

  // Human-readable message describing the error. Shouldn't be used for
  // anything but logging/diagnostics, since messages are not guaranteed to be
  // stable.
  const char* message() const;
  
  int32_t error_code() const;
  
  int32_t error_code_ex() const;
  
  void set_error_code(int32_t errorCode);
  
  void set_error_code_ex(int32_t errorCode, int32_t errorCodeExtend);
  
  void set_message(std::string message);

  // Convenience method for situations where you only care whether or not an
  // error occurred.
  bool ok() const { return type_ == RTCErrorType::NONE; }

 private:
  RTCErrorType type_ = RTCErrorType::NONE;
  
  std::string message_;
  
  int32_t _errorCode;
  int32_t _errorCodeExtend; //扩展错误码
};

const char* ToString(RTCErrorType error);

}
#endif /* RTC_ERROR_H */
